# Copyright 1998-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# pylint: disable=ungrouped-imports

# ===========================================================================
# START OF IMPORTS -- START OF IMPORTS -- START OF IMPORTS -- START OF IMPORT
# ===========================================================================

from portage import installation

try:
    import asyncio
    import sys
    import errno

    if not hasattr(errno, "ESTALE"):
        # ESTALE may not be defined on some systems, such as interix.
        errno.ESTALE = -1
    import multiprocessing.util
    import re
    import types
    import platform

    # Temporarily delete these imports, to ensure that only the
    # wrapped versions are imported by portage internals.
    import os

    del os
    import shutil

    del shutil

except ImportError as e:
    sys.stderr.write("\n\n")
    sys.stderr.write(
        "!!! Failed to complete python imports. These are internal modules for\n"
    )
    sys.stderr.write(
        "!!! python and failure here indicates that you have a problem with python\n"
    )
    sys.stderr.write(
        "!!! itself and thus portage is not able to continue processing.\n\n"
    )

    sys.stderr.write(
        "!!! You might consider starting python with verbose flags to see what has\n"
    )
    sys.stderr.write(
        "!!! gone wrong. Here is the information we got for this exception:\n"
    )
    sys.stderr.write(f"    {e}\n\n")
    raise

try:
    import portage.proxy.lazyimport
    import portage.proxy as proxy

    proxy.lazyimport.lazyimport(
        globals(),
        "portage.cache.cache_errors:CacheError",
        "portage.checksum",
        "portage.checksum:perform_checksum,perform_md5,prelink_capable",
        "portage.cvstree",
        "portage.data",
        "portage.data:lchown,ostype,portage_gid,portage_uid,secpass,"
        + "uid,userland,userpriv_groups,wheelgid",
        "portage.dbapi",
        "portage.dbapi.bintree:bindbapi,binarytree",
        "portage.dbapi.cpv_expand:cpv_expand",
        "portage.dbapi.dep_expand:dep_expand",
        "portage.dbapi.porttree:close_portdbapi_caches,FetchlistDict,"
        + "portagetree,portdbapi",
        "portage.dbapi.vartree:dblink,merge,unmerge,vardbapi,vartree",
        "portage.dbapi.virtual:fakedbapi",
        "portage.debug",
        "portage.dep",
        "portage.dep:best_match_to_list,dep_getcpv,dep_getkey,"
        + "flatten,get_operator,isjustname,isspecific,isvalidatom,"
        + "match_from_list,match_to_list",
        "portage.dep.dep_check:dep_check,dep_eval,dep_wordreduce,dep_zapdeps",
        "portage.eclass_cache",
        "portage.elog",
        "portage.exception",
        "portage.getbinpkg",
        "portage.locks",
        "portage.locks:lockdir,lockfile,unlockdir,unlockfile",
        "portage.mail",
        "portage.manifest:Manifest",
        "portage.output",
        "portage.output:bold,colorize",
        "portage.package.ebuild.doebuild:doebuild,"
        + "doebuild_environment,spawn,spawnebuild",
        "portage.package.ebuild.config:autouse,best_from_dict,"
        + "check_config_instance,config",
        "portage.package.ebuild.deprecated_profile_check:deprecated_profile_check",
        "portage.package.ebuild.digestcheck:digestcheck",
        "portage.package.ebuild.digestgen:digestgen",
        "portage.package.ebuild.fetch:fetch",
        "portage.package.ebuild.getmaskingreason:getmaskingreason",
        "portage.package.ebuild.getmaskingstatus:getmaskingstatus",
        "portage.package.ebuild.prepare_build_dirs:prepare_build_dirs",
        "portage.process",
        "portage.process:atexit_register,run_exitfuncs",
        "portage.update:dep_transform,fixdbentries,grab_updates,"
        + "parse_updates,update_config_files,update_dbentries,"
        + "update_dbentry",
        "portage.util",
        "portage.util:atomic_ofstream,apply_secpass_permissions,"
        + "apply_recursive_permissions,dump_traceback,getconfig,"
        + "grabdict,grabdict_package,grabfile,grabfile_package,"
        + "map_dictlist_vals,new_protect_filename,normalize_path,"
        + "pickle_read,pickle_write,stack_dictlist,stack_dicts,"
        + "stack_lists,unique_array,varexpand,writedict,writemsg,"
        + "writemsg_stdout,write_atomic",
        "portage.util.digraph:digraph",
        "portage.util.env_update:env_update",
        "portage.util.ExtractKernelVersion:ExtractKernelVersion",
        "portage.util.listdir:cacheddir,listdir",
        "portage.util.movefile:movefile",
        "portage.util.mtimedb:MtimeDB",
        "portage.versions",
        "portage.versions:best,catpkgsplit,catsplit,cpv_getkey,"
        + "cpv_getkey@getCPFromCPV,endversion_keys,"
        + "suffix_value@endversion,pkgcmp,pkgsplit,vercmp,ververify",
        "portage.xpak",
        "portage.gpkg",
        "subprocess",
        "time",
    )

    from collections import OrderedDict

    import portage.const
    from portage.const import (
        VDB_PATH,
        PRIVATE_PATH,
        CACHE_PATH,
        DEPCACHE_PATH,
        USER_CONFIG_PATH,
        MODULES_FILE_PATH,
        CUSTOM_PROFILE_PATH,
        PORTAGE_BASE_PATH,
        PORTAGE_BIN_PATH,
        PORTAGE_PYM_PATH,
        PROFILE_PATH,
        LOCALE_DATA_PATH,
        EBUILD_SH_BINARY,
        SANDBOX_BINARY,
        BASH_BINARY,
        MOVE_BINARY,
        PRELINK_BINARY,
        WORLD_FILE,
        MAKE_CONF_FILE,
        MAKE_DEFAULTS_FILE,
        DEPRECATED_PROFILE_FILE,
        USER_VIRTUALS_FILE,
        EBUILD_SH_ENV_FILE,
        INVALID_ENV_FILE,
        CUSTOM_MIRRORS_FILE,
        CONFIG_MEMORY_FILE,
        INCREMENTALS,
        EAPI,
        MISC_SH_BINARY,
        REPO_NAME_LOC,
        REPO_NAME_FILE,
    )

except ImportError as e:
    sys.stderr.write("\n\n")
    sys.stderr.write(
        "!!! Failed to complete portage imports. There are internal modules for\n"
    )
    sys.stderr.write(
        "!!! portage and failure here indicates that you have a problem with your\n"
    )
    sys.stderr.write(
        "!!! installation of portage. Please try a rescue portage located in the ebuild\n"
    )
    sys.stderr.write(
        "!!! repository under '/var/db/repos/gentoo/sys-apps/portage/files/' (default).\n"
    )
    sys.stderr.write(
        "!!! There is a README.RESCUE file that details the steps required to perform\n"
    )
    sys.stderr.write("!!! a recovery of portage.\n")
    sys.stderr.write(f"    {e}\n\n")
    raise


# We use utf_8 encoding everywhere. Previously, we used
# sys.getfilesystemencoding() for the 'merge' encoding, but that had
# various problems:
#
#   1) If the locale is ever changed then it can cause orphan files due
#      to changed character set translation.
#
#   2) Ebuilds typically install files with utf_8 encoded file names,
#      and then portage would be forced to rename those files to match
#      sys.getfilesystemencoding(), possibly breaking things.
#
#   3) Automatic translation between encodings can lead to nonsensical
#      file names when the source encoding is unknown by portage.
#
#   4) It's inconvenient for ebuilds to convert the encodings of file
#      names to match the current locale, and upstreams typically encode
#      file names with utf_8 encoding.
#
# So, instead of relying on sys.getfilesystemencoding(), we avoid the above
# problems by using a constant utf_8 'merge' encoding for all locales, as
# discussed in bug #382199 and bug #381509.
_encodings = {
    "content": "utf_8",
    "fs": "utf_8",
    "merge": "utf_8",
    "repo.content": "utf_8",
    "stdio": "utf_8",
}


def _decode_argv(argv):
    # With Python 3, the surrogateescape encoding error handler makes it
    # possible to access the original argv bytes, which can be useful
    # if their actual encoding does no match the filesystem encoding.
    fs_encoding = sys.getfilesystemencoding()
    return [_unicode_decode(x.encode(fs_encoding, "surrogateescape")) for x in argv]


def _unicode_encode(s, encoding=_encodings["content"], errors="backslashreplace"):
    if isinstance(s, str):
        s = s.encode(encoding, errors)
    return s


def _unicode_decode(s, encoding=_encodings["content"], errors="replace"):
    if isinstance(s, bytes):
        s = str(s, encoding=encoding, errors=errors)
    return s


_native_string = _unicode_decode


class _unicode_func_wrapper:
    """
    Wraps a function, converts arguments from unicode to bytes,
    and return values to unicode from bytes. Function calls
    will raise UnicodeEncodeError if an argument fails to be
    encoded with the required encoding. Return values that
    are single strings are decoded with errors='replace'. Return
    values that are lists of strings are decoded with errors='strict'
    and elements that fail to be decoded are omitted from the returned
    list.
    """

    __slots__ = ("_func", "_encoding")

    def __init__(self, func, encoding=_encodings["fs"]):
        self._func = func
        self._encoding = encoding

    def _process_args(self, args, kwargs):
        encoding = self._encoding
        wrapped_args = [
            _unicode_encode(x, encoding=encoding, errors="strict") for x in args
        ]
        if kwargs:
            wrapped_kwargs = {
                k: _unicode_encode(v, encoding=encoding, errors="strict")
                for k, v in kwargs.items()
            }
        else:
            wrapped_kwargs = {}

        return (wrapped_args, wrapped_kwargs)

    def __call__(self, *args, **kwargs):
        encoding = self._encoding
        wrapped_args, wrapped_kwargs = self._process_args(args, kwargs)

        rval = self._func(*wrapped_args, **wrapped_kwargs)

        # Don't use isinstance() since we don't want to convert subclasses
        # of tuple such as posix.stat_result in Python >=3.2.
        if rval.__class__ in (list, tuple):
            decoded_rval = []
            for x in rval:
                try:
                    x = _unicode_decode(x, encoding=encoding, errors="strict")
                except UnicodeDecodeError:
                    pass
                else:
                    decoded_rval.append(x)

            if isinstance(rval, tuple):
                rval = tuple(decoded_rval)
            else:
                rval = decoded_rval
        else:
            rval = _unicode_decode(rval, encoding=encoding, errors="replace")

        return rval


class _unicode_module_wrapper:
    """
    Wraps a module and wraps all functions with _unicode_func_wrapper.
    """

    __slots__ = ("_mod", "_encoding", "_overrides", "_cache")

    def __init__(self, mod, encoding=_encodings["fs"], overrides=None, cache=True):
        object.__setattr__(self, "_mod", mod)
        object.__setattr__(self, "_encoding", encoding)
        object.__setattr__(self, "_overrides", overrides)
        if cache:
            cache = {}
        else:
            cache = None
        object.__setattr__(self, "_cache", cache)

    def __getattribute__(self, attr):
        cache = object.__getattribute__(self, "_cache")
        if cache is not None:
            result = cache.get(attr)
            if result is not None:
                return result
        result = getattr(object.__getattribute__(self, "_mod"), attr)
        encoding = object.__getattribute__(self, "_encoding")
        overrides = object.__getattribute__(self, "_overrides")
        override = None
        if overrides is not None:
            override = overrides.get(id(result))
        if override is not None:
            result = override
        elif isinstance(result, type):
            pass
        elif type(result) is types.ModuleType:
            result = _unicode_module_wrapper(
                result, encoding=encoding, overrides=overrides
            )
        elif hasattr(result, "__call__"):
            result = _unicode_func_wrapper(result, encoding=encoding)
        if cache is not None:
            cache[attr] = result
        return result


class _eintr_func_wrapper:
    """
    Wraps a function and handles EINTR by calling the function as
    many times as necessary (until it returns without raising EINTR).
    """

    __slots__ = ("_func",)

    def __init__(self, func):
        self._func = func

    def __call__(self, *args, **kwargs):
        while True:
            try:
                rval = self._func(*args, **kwargs)
                break
            except OSError as e:
                if e.errno != errno.EINTR:
                    raise

        return rval


import os as _os

_os_overrides = {
    id(_os.fdopen): _os.fdopen,
    id(_os.popen): _os.popen,
    id(_os.read): _os.read,
    id(_os.system): _os.system,
    id(_os.waitpid): _eintr_func_wrapper(_os.waitpid),
}


_os_overrides[id(_os.mkfifo)] = _os.mkfifo

if hasattr(_os, "statvfs"):
    _os_overrides[id(_os.statvfs)] = _os.statvfs

os = _unicode_module_wrapper(_os, overrides=_os_overrides, encoding=_encodings["fs"])
_os_merge = _unicode_module_wrapper(
    _os, encoding=_encodings["merge"], overrides=_os_overrides
)

import shutil as _shutil

shutil = _unicode_module_wrapper(_shutil, encoding=_encodings["fs"])

# Imports below this point rely on the above unicode wrapper definitions.
try:
    __import__("selinux")
    import portage._selinux

    selinux = _unicode_module_wrapper(_selinux, encoding=_encodings["fs"])
    _selinux_merge = _unicode_module_wrapper(_selinux, encoding=_encodings["merge"])
except (ImportError, OSError) as e:
    if isinstance(e, OSError):
        sys.stderr.write(f"!!! SELinux not loaded: {e}\n")
    del e
    _selinux = None
    selinux = None
    _selinux_merge = None

# ===========================================================================
# END OF IMPORTS -- END OF IMPORTS -- END OF IMPORTS -- END OF IMPORTS -- END
# ===========================================================================

_python_interpreter = (
    sys.executable
    if os.environ.get("VIRTUAL_ENV")
    else os.path.realpath(sys.executable)
)
_bin_path = PORTAGE_BIN_PATH
_pym_path = PORTAGE_PYM_PATH
_not_installed = os.path.isfile(
    os.path.join(PORTAGE_BASE_PATH, ".portage_not_installed")
)

# Api consumers included in portage should set this to True.
_internal_caller = False

_sync_mode = False


class _ForkWatcher:
    @staticmethod
    def hook(_ForkWatcher):
        _ForkWatcher.current_pid = None
        # Force instantiation of a new event loop policy as a workaround
        # for https://bugs.python.org/issue22087.
        asyncio.set_event_loop_policy(None)


_ForkWatcher.hook(_ForkWatcher)

multiprocessing.util.register_after_fork(_ForkWatcher, _ForkWatcher.hook)


def getpid():
    """
    Cached version of os.getpid(). ForkProcess updates the cache.
    """
    if _ForkWatcher.current_pid is None:
        _ForkWatcher.current_pid = _os.getpid()
    return _ForkWatcher.current_pid


def _get_stdin():
    """
    Buggy code in python's multiprocessing/process.py closes sys.stdin
    and reassigns it to open(os.devnull), but fails to update the
    corresponding __stdin__ reference. So, detect that case and handle
    it appropriately.
    """
    if not sys.__stdin__.closed:
        return sys.__stdin__
    return sys.stdin


_shell_quote_re = re.compile(r"[\s><=*\\\"'$`;&|(){}\[\]#!~?]")


def _shell_quote(s):
    """
    Quote a string in double-quotes and use backslashes to
    escape any backslashes, double-quotes, dollar signs, or
    backquotes in the string.
    """
    if _shell_quote_re.search(s) is None:
        return s
    for letter in r"\"$`":
        if letter in s:
            s = s.replace(letter, rf"\{letter}")
    return f'"{s}"'


bsd_chflags = None

if platform.system() in ("FreeBSD",):
    # TODO: remove this class?
    class bsd_chflags:
        chflags = os.chflags
        lchflags = os.lchflags


def load_mod(name):
    components = name.split(".")
    modname = ".".join(components[:-1])
    mod = __import__(modname)
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def getcwd():
    "this fixes situations where the current directory doesn't exist"
    try:
        return os.getcwd()
    except OSError:  # dir doesn't exist
        os.chdir("/")
        return "/"


getcwd()


def abssymlink(symlink, target=None):
    """
    This reads symlinks, resolving the relative symlinks,
    and returning the absolute.
    @param symlink: path of symlink (must be absolute)
    @param target: the target of the symlink (as returned
            by readlink)
    @rtype: str
    @return: the absolute path of the symlink target
    """
    if target is not None:
        mylink = target
    else:
        mylink = os.readlink(symlink)
    if mylink[0] != "/":
        mydir = os.path.dirname(symlink)
        mylink = f"{mydir}/{mylink}"
    return os.path.normpath(mylink)


_doebuild_manifest_exempt_depend = 0

_testing_eapis = frozenset([])
_deprecated_eapis = frozenset(
    [
        "3_pre1",
        "3_pre2",
        "4_pre1",
        "4-slot-abi",
        "5_pre1",
        "5_pre2",
        "6_pre1",
        "7_pre1",
    ]
)

from itertools import chain

_supported_eapis = frozenset(
    chain(
        (str(x) for x in range(portage.const.EAPI + 1)),
        _testing_eapis,
        _deprecated_eapis,
    )
)


def _eapi_is_deprecated(eapi):
    return eapi in _deprecated_eapis


def eapi_is_supported(eapi):
    eapi = str(eapi).strip()
    return eapi in _supported_eapis


# This pattern is specified by PMS section 7.3.1.
_pms_eapi_re = re.compile(r"^[ \t]*EAPI=(['\"]?)([A-Za-z0-9+_.-]*)\1[ \t]*([ \t]#.*)?$")
_comment_or_blank_line = re.compile(r"^\s*(#.*)?$")


def _parse_eapi_ebuild_head(f):
    eapi = None
    eapi_lineno = None
    lineno = 0
    for line in f:
        lineno += 1
        m = _comment_or_blank_line.match(line)
        if m is None:
            eapi_lineno = lineno
            m = _pms_eapi_re.match(line)
            if m is not None:
                eapi = m.group(2)
            break

    return (eapi, eapi_lineno)


def _movefile(src, dest, **kwargs):
    """Calls movefile and raises a PortageException if an error occurs."""
    if movefile(src, dest, **kwargs) is None:
        raise portage.exception.PortageException(f"mv '{src}' '{dest}'")


auxdbkeys = (
    "DEPEND",
    "RDEPEND",
    "SLOT",
    "SRC_URI",
    "RESTRICT",
    "HOMEPAGE",
    "LICENSE",
    "DESCRIPTION",
    "KEYWORDS",
    "INHERITED",
    "IUSE",
    "REQUIRED_USE",
    "PDEPEND",
    "BDEPEND",
    "EAPI",
    "PROPERTIES",
    "DEFINED_PHASES",
    "IDEPEND",
    "INHERIT",
)


def portageexit():
    pass


class _trees_dict(dict):
    __slots__ = (
        "_running_eroot",
        "_target_eroot",
    )

    def __init__(self, *pargs, **kargs):
        dict.__init__(self, *pargs, **kargs)
        self._running_eroot = None
        self._target_eroot = None


def create_trees(
    config_root=None, target_root=None, trees=None, env=None, sysroot=None, eprefix=None
):
    if trees is None:
        trees = _trees_dict()
    elif not isinstance(trees, _trees_dict):
        # caller passed a normal dict or something,
        # but we need a _trees_dict instance
        trees = _trees_dict(trees)

    if env is None:
        env = os.environ

    settings = config(
        config_root=config_root,
        target_root=target_root,
        env=env,
        sysroot=sysroot,
        eprefix=eprefix,
    )
    settings.lock()

    depcachedir = settings.get("PORTAGE_DEPCACHEDIR")
    trees._target_eroot = settings["EROOT"]
    myroots = [(settings["EROOT"], settings)]
    if settings["ROOT"] == "/" and settings["EPREFIX"] == const.EPREFIX:
        trees._running_eroot = trees._target_eroot
    else:
        # When ROOT != "/" we only want overrides from the calling
        # environment to apply to the config that's associated
        # with ROOT != "/", so pass a nearly empty dict for the env parameter.
        env_sequence = (
            "PATH",
            "PORTAGE_GRPNAME",
            "PORTAGE_REPOSITORIES",
            "PORTAGE_USERNAME",
            "PYTHONPATH",
            "SSH_AGENT_PID",
            "SSH_AUTH_SOCK",
            "TERM",
            "ftp_proxy",
            "http_proxy",
            "https_proxy",
            "no_proxy",
            "__PORTAGE_TEST_HARDLINK_LOCKS",
        )
        env = ((k, settings.get(k)) for k in env_sequence)
        clean_env = {k: v for k, v in env if v is not None}

        if depcachedir is not None:
            clean_env["PORTAGE_DEPCACHEDIR"] = depcachedir
        mysettings = config(
            config_root=None, target_root="/", env=clean_env, sysroot="/", eprefix=None
        )
        mysettings.lock()
        trees._running_eroot = mysettings["EROOT"]
        myroots.append((mysettings["EROOT"], mysettings))

        if settings["SYSROOT"] != "/" and settings["SYSROOT"] != settings["ROOT"]:
            mysettings = config(
                config_root=settings["SYSROOT"],
                target_root=settings["SYSROOT"],
                env=clean_env,
                sysroot=settings["SYSROOT"],
                eprefix="",
            )
            mysettings.lock()
            myroots.append((mysettings["EROOT"], mysettings))

    for myroot, mysettings in myroots:
        trees[myroot] = portage.util.LazyItemsDict(trees.get(myroot, {}))
        trees[myroot].addLazySingleton("virtuals", mysettings.getvirtuals)
        trees[myroot].addLazySingleton(
            "vartree", vartree, categories=mysettings.categories, settings=mysettings
        )
        trees[myroot].addLazySingleton("porttree", portagetree, settings=mysettings)
        trees[myroot].addLazySingleton(
            "bintree", binarytree, pkgdir=mysettings["PKGDIR"], settings=mysettings
        )
    return trees


if installation.TYPE == installation.TYPES.SOURCE:

    class _LazyVersion(proxy.objectproxy.ObjectProxy):
        def _get_target(self):
            global VERSION
            if VERSION is not self:
                return VERSION
            if os.path.isdir(os.path.join(PORTAGE_BASE_PATH, ".git")):
                encoding = _encodings["fs"]
                cmd = [
                    BASH_BINARY,
                    "-c",
                    (
                        f"cd {_shell_quote(PORTAGE_BASE_PATH)} ; git describe --match 'portage-*' || exit $? ; "
                        'if [ -n "`git diff-index --name-only --diff-filter=M HEAD`" ] ; '
                        "then echo modified ; git rev-list --format=%%ct -n 1 HEAD ; fi ; "
                        "exit 0"
                    ),
                ]
                cmd = [
                    _unicode_encode(x, encoding=encoding, errors="strict") for x in cmd
                ]
                proc = subprocess.Popen(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                )
                output = _unicode_decode(proc.communicate()[0], encoding=encoding)
                status = proc.wait()
                if os.WIFEXITED(status) and os.WEXITSTATUS(status) == os.EX_OK:
                    output_lines = output.splitlines()
                    if output_lines:
                        version_split = output_lines[0].split("-")
                        if len(version_split) > 1:
                            VERSION = version_split[1]
                            patchlevel = False
                            if len(version_split) > 2:
                                patchlevel = True
                                VERSION = f"{VERSION}_p{version_split[2]}"
                            if len(output_lines) > 1 and output_lines[1] == "modified":
                                head_timestamp = None
                                if len(output_lines) > 3:
                                    try:
                                        head_timestamp = int(output_lines[3])
                                    except ValueError:
                                        pass
                                timestamp = int(time.time())
                                if (
                                    head_timestamp is not None
                                    and timestamp > head_timestamp
                                ):
                                    timestamp = timestamp - head_timestamp
                                if not patchlevel:
                                    VERSION = f"{VERSION}_p0"
                                VERSION = f"{VERSION}_p{timestamp}"
                            return VERSION
            VERSION = "HEAD"
            return VERSION

    VERSION = _LazyVersion()

else:
    VERSION = "@VERSION@"

_legacy_global_var_names = (
    "archlist",
    "db",
    "features",
    "groups",
    "mtimedb",
    "mtimedbfile",
    "pkglines",
    "portdb",
    "profiledir",
    "root",
    "selinux_enabled",
    "settings",
    "thirdpartymirrors",
)


def _reset_legacy_globals():
    global _legacy_globals_constructed
    _legacy_globals_constructed = set()
    for k in _legacy_global_var_names:
        globals()[k] = _LegacyGlobalProxy(k)


class _LegacyGlobalProxy(proxy.objectproxy.ObjectProxy):
    __slots__ = ("_name",)

    def __init__(self, name):
        proxy.objectproxy.ObjectProxy.__init__(self)
        object.__setattr__(self, "_name", name)

    def _get_target(self):
        name = object.__getattribute__(self, "_name")
        from portage._legacy_globals import _get_legacy_global

        return _get_legacy_global(name)


_reset_legacy_globals()


def _disable_legacy_globals():
    """
    This deletes the ObjectProxy instances that are used
    for lazy initialization of legacy global variables.
    The purpose of deleting them is to prevent new code
    from referencing these deprecated variables.
    """
    global _legacy_global_var_names
    for k in _legacy_global_var_names:
        globals().pop(k, None)
    portage.data._initialized_globals.clear()



# Copyright 2005-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.exception import PortageException
from portage.cache.mappings import ProtectedDict
from portage.localization import _
from portage.util import writemsg


class InvalidModuleName(PortageException):
    """An invalid or unknown module name."""


class ModuleVersionError(PortageException):
    """An incompatible module version"""


class Module:
    """Class to define and hold our plug-in module

    @type name: string
    @param name: the module name
    @type path: the path to the new module
    """

    def __init__(self, name, namepath):
        """Some variables initialization"""
        self.name = name
        self._namepath = namepath
        self.kids_names = []
        self.kids = {}
        self.initialized = self._initialize()

    def _initialize(self):
        """Initialize the plug-in module

        @rtype: boolean
        """
        self.valid = False
        try:
            mod_name = ".".join([self._namepath, self.name])
            self._module = __import__(mod_name, [], [], ["not empty"])
            self.valid = True
        except ImportError as e:
            print("MODULE; failed import", mod_name, "  error was:", e)
            return False
        self.module_spec = self._module.module_spec
        for submodule in self.module_spec["provides"]:
            kid = self.module_spec["provides"][submodule]
            kidname = kid["name"]
            try:
                kid["module_name"] = ".".join([mod_name, kid["sourcefile"]])
            except KeyError:
                kid["module_name"] = ".".join([mod_name, self.name])
                writemsg(
                    _(
                        f"{self.name} module's module_spec is old, missing attribute: "
                        "'sourcefile'.  Backward compatibility may be "
                        f"removed in the future.\nFile: {self._module.__file__}\n"
                    )
                )
            kid["is_imported"] = False
            self.kids[kidname] = kid
            self.kids_names.append(kidname)
        return True

    def get_class(self, name):
        if not name or name not in self.kids_names:
            raise InvalidModuleName(
                f"Module name '{name}' is invalid or not"
                f"part of the module '{self.name}'"
            )
        kid = self.kids[name]
        if kid["is_imported"]:
            module = kid["instance"]
        else:
            try:
                module = __import__(kid["module_name"], [], [], ["not empty"])
                kid["instance"] = module
                kid["is_imported"] = True
            except ImportError:
                raise
        mod_class = getattr(module, kid["class"])
        return mod_class


class Modules:
    """Dynamic modules system for loading and retrieving any of the
    installed emaint modules and/or provided class's

    @param path: Path to the "modules" directory
    @param namepath: Python import path to the "modules" directory
    """

    def __init__(self, path, namepath, compat_versions=None):
        self._module_path = path
        self._namepath = namepath
        self.compat_versions = compat_versions
        self.parents = []
        self._modules = self._get_all_modules()
        self.modules = ProtectedDict(self._modules)
        self.module_names = sorted(self._modules)

    def _get_all_modules(self):
        """scans the _module_path dir for loadable modules

        @rtype: dictionary of module_plugins
        """
        module_dir = self._module_path
        names = os.listdir(module_dir)

        def _a_real_module(entry):
            try:
                # test for statinfo to ensure it should a real module
                # it will bail if it errors
                os.lstat(os.path.join(module_dir, entry, "__init__.py"))
            except OSError:
                return False
            return True

        # The importables list cannot be a generator.
        # If it was a generator, it would be consumed by self.parents.extend()
        # and the following for loop wouldn't have anything to iterate with.
        importables = [
            entry
            for entry in names
            if not entry.startswith("__") and _a_real_module(entry)
        ]
        self.parents.extend(importables)

        kids = {}
        for entry in importables:
            new_module = Module(entry, self._namepath)
            self._check_compat(new_module)
            for module_name in new_module.kids:
                kid = new_module.kids[module_name]
                kid["parent"] = new_module
                kids[kid["name"]] = kid
        return kids

    def get_module_names(self):
        """Convenience function to return the list of installed modules
        available

        @rtype: list
        @return: the installed module names available
        """
        return self.module_names

    def get_class(self, modname):
        """Retrieves a module class desired

        @type modname: string
        @param modname: the module class name
        """
        if modname and modname in self.module_names:
            mod = self._modules[modname]["parent"].get_class(modname)
        else:
            raise InvalidModuleName(f"Module name '{modname}' is invalid or not found")
        return mod

    def get_description(self, modname):
        """Retrieves the module class decription

        @type modname: string
        @param modname: the module class name
        @type string
        @return: the modules class decription
        """
        if modname and modname in self.module_names:
            mod = self._modules[modname]["description"]
        else:
            raise InvalidModuleName(f"Module name '{modname}' is invalid or not found")
        return mod

    def get_functions(self, modname):
        """Retrieves the module class  exported function names

        @type modname: string
        @param modname: the module class name
        @type list
        @return: the modules class exported function names
        """
        if modname and modname in self.module_names:
            mod = self._modules[modname]["functions"]
        else:
            raise InvalidModuleName(f"Module name '{modname}' is invalid or not found")
        return mod

    def get_func_descriptions(self, modname):
        """Retrieves the module class  exported functions descriptions

        @type modname: string
        @param modname: the module class name
        @type dictionary
        @return: the modules class exported functions descriptions
        """
        if modname and modname in self.module_names:
            desc = self._modules[modname]["func_desc"]
        else:
            raise InvalidModuleName(f"Module name '{modname}' is invalid or not found")
        return desc

    def get_opt_descriptions(self, modname):
        """Retrieves the module class exported options descriptions

        @type modname: string
        @param modname: the module class name
        @type dictionary
        @return: the modules class exported options descriptions
        """
        if modname and modname in self.module_names:
            desc = self._modules[modname].get("opt_desc")
        else:
            raise InvalidModuleName(f"Module name '{modname}' is invalid or not found")
        return desc

    def get_spec(self, modname, var):
        """Retrieves the module class exported spec variable

        @type modname: string
        @param modname: the module class name
        @type var: string
        @param var: the base level variable to return
        @type dictionary
        @return: the modules class exported options descriptions
        """
        if modname and modname in self.module_names:
            value = self._modules[modname].get(var, None)
        else:
            raise InvalidModuleName(f"Module name '{modname}' is invalid or not found")
        return value

    def _check_compat(self, module):
        if self.compat_versions:
            if not module.module_spec["version"] in self.compat_versions:
                raise ModuleVersionError(
                    f"Error loading '{self._namepath}' plugin module: {module.module_spec['name']}, version: {module.module_spec['version']}\n"
                    "Module is not compatible with the current application version\n"
                    f"Compatible module API versions are: {self.compat_versions}"
                )



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import time
import portage
from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.data import secpass
from portage.output import xtermTitle

# We disable emergelog by default, since it's called from
# dblink.merge() and we don't want that to trigger log writes
# unless it's really called via emerge.
_disable = True
_emerge_log_dir = "/var/log"


def emergelog(xterm_titles, mystr, short_msg=None):
    if _disable:
        return

    mystr = _unicode_decode(mystr)

    if short_msg is not None:
        short_msg = _unicode_decode(short_msg)

    if xterm_titles and short_msg:
        if "HOSTNAME" in os.environ:
            short_msg = os.environ["HOSTNAME"] + ": " + short_msg
        xtermTitle(short_msg)
    try:
        file_path = os.path.join(_emerge_log_dir, "emerge.log")
        existing_log = os.path.exists(file_path)
        mylogfile = open(
            _unicode_encode(file_path, encoding=_encodings["fs"], errors="strict"),
            mode="a",
            encoding=_encodings["content"],
            errors="backslashreplace",
        )
        if not existing_log:
            portage.util.apply_secpass_permissions(
                file_path, uid=portage.portage_uid, gid=portage.portage_gid, mode=0o660
            )
        mylock = portage.locks.lockfile(file_path)
        try:
            mylogfile.write(f"{time.time():.0f}: {mystr}\n")
            mylogfile.close()
        finally:
            portage.locks.unlockfile(mylock)
    except (OSError, portage.exception.PortageException) as e:
        if secpass >= 1:
            portage.util.writemsg(f"emergelog(): {e}\n", noiselevel=-1)



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import stat

from portage import best, os
from portage.const import WORLD_FILE
from portage.data import secpass
from portage.exception import DirectoryNotFound
from portage.localization import _
from portage.output import bold, colorize
from portage.update import (
    grab_updates,
    parse_updates,
    update_config_files,
    update_dbentry,
)
from portage.util import grabfile, shlex_split, writemsg, writemsg_stdout, write_atomic


def _global_updates(trees, prev_mtimes, quiet=False, if_mtime_changed=True):
    """
    Perform new global updates if they exist in 'profiles/updates/'
    subdirectories of all active repositories (PORTDIR + PORTDIR_OVERLAY).
    This simply returns if ROOT != "/" (when len(trees) != 1). If ROOT != "/"
    then the user should instead use emaint --fix movebin and/or moveinst.

    @param trees: A dictionary containing package databases.
    @type trees: dict
    @param prev_mtimes: A dictionary containing mtimes of files located in
            $PORTDIR/profiles/updates/.
    @type prev_mtimes: dict
    @rtype: bool
    @return: True if update commands have been performed, otherwise False
    """
    # only do this if we're root and not running repoman/ebuild digest

    if secpass < 2 or "SANDBOX_ACTIVE" in os.environ or len(trees) != 1:
        return False

    return _do_global_updates(
        trees, prev_mtimes, quiet=quiet, if_mtime_changed=if_mtime_changed
    )


def _do_global_updates(trees, prev_mtimes, quiet=False, if_mtime_changed=True):
    root = trees._running_eroot
    mysettings = trees[root]["vartree"].settings
    portdb = trees[root]["porttree"].dbapi
    vardb = trees[root]["vartree"].dbapi
    bindb = trees[root]["bintree"].dbapi

    world_file = os.path.join(mysettings["EROOT"], WORLD_FILE)
    world_list = grabfile(world_file)
    world_modified = False
    world_warnings = set()
    updpath_map = {}
    # Maps repo_name to list of updates. If a given repo has no updates
    # directory, it will be omitted. If a repo has an updates directory
    # but none need to be applied (according to timestamp logic), the
    # value in the dict will be an empty list.
    repo_map = {}
    timestamps = {}

    retupd = False
    update_notice_printed = False
    for repo_name in portdb.getRepositories():
        repo = portdb.getRepositoryPath(repo_name)
        updpath = os.path.join(repo, "profiles", "updates")
        if not os.path.isdir(updpath):
            continue

        if updpath in updpath_map:
            repo_map[repo_name] = updpath_map[updpath]
            continue

        try:
            if if_mtime_changed:
                update_data = grab_updates(updpath, prev_mtimes=prev_mtimes)
            else:
                update_data = grab_updates(updpath)
        except DirectoryNotFound:
            continue
        myupd = []
        updpath_map[updpath] = myupd
        repo_map[repo_name] = myupd
        if len(update_data) > 0:
            for mykey, mystat, mycontent in update_data:
                if not update_notice_printed:
                    update_notice_printed = True
                    writemsg_stdout("\n")
                    writemsg_stdout(colorize("GOOD", "Performing Global Updates\n"))
                    writemsg_stdout(
                        _(
                            "(Could take a couple of minutes if you have a lot of binary packages.)\n"
                        )
                    )
                    if not quiet:
                        writemsg_stdout(
                            "  ".join(
                                (
                                    "",
                                    f"{bold('.')}='update pass'",
                                    f"{bold('*')}='binary update'",
                                    f"{bold('#')}='/var/db update'",
                                    f"{bold('@')}='/var/db move'\n",
                                    f"{bold('s')}='/var/db SLOT move'",
                                    f"{bold('%')}='binary move'",
                                    f"{bold('S')}='binary SLOT move'\n",
                                    f"{bold('p')}='update /etc/portage/package.*'\n",
                                )
                            )
                        )
                valid_updates, errors = parse_updates(mycontent)
                myupd.extend(valid_updates)
                if not quiet:
                    writemsg_stdout(bold(mykey))
                    writemsg_stdout(f"{len(valid_updates) * '.'}\n")
                if len(errors) == 0:
                    # Update our internal mtime since we
                    # processed all of our directives.
                    timestamps[mykey] = mystat[stat.ST_MTIME]
                else:
                    for msg in errors:
                        writemsg(f"{msg}\n", noiselevel=-1)
            if myupd:
                retupd = True

    if retupd:
        if os.access(bindb.bintree.pkgdir, os.W_OK):
            # Call binarytree.populate(), since we want to make sure it's
            # only populated with local packages here (getbinpkgs=0).
            bindb.bintree.populate()
        else:
            bindb = None

    master_repo = portdb.repositories.mainRepo()
    if master_repo is not None:
        master_repo = master_repo.name
    if master_repo in repo_map:
        repo_map["DEFAULT"] = repo_map[master_repo]

    for repo_name, myupd in repo_map.items():
        if repo_name == "DEFAULT":
            continue
        if not myupd:
            continue

        def repo_match(repository):
            return repository == repo_name or (
                repo_name == master_repo and repository not in repo_map
            )

        def _world_repo_match(atoma, atomb):
            """
            Check whether to perform a world change from atoma to atomb.
            If best vardb match for atoma comes from the same repository
            as the update file, allow that. Additionally, if portdb still
            can find a match for old atom name, warn about that.
            """
            matches = vardb.match(atoma)
            if not matches:
                matches = vardb.match(atomb)
            if matches and repo_match(vardb.aux_get(best(matches), ["repository"])[0]):
                if portdb.match(atoma):
                    world_warnings.add((atoma, atomb))
                return True
            return False

        for update_cmd in myupd:
            for pos, atom in enumerate(world_list):
                new_atom = update_dbentry(update_cmd, atom)
                if atom != new_atom:
                    if _world_repo_match(atom, new_atom):
                        world_list[pos] = new_atom
                        world_modified = True

        for update_cmd in myupd:
            if update_cmd[0] == "move":
                moves = vardb.move_ent(update_cmd, repo_match=repo_match)
                if moves:
                    writemsg_stdout(moves * "@")
                if bindb:
                    moves = bindb.move_ent(update_cmd, repo_match=repo_match)
                    if moves:
                        writemsg_stdout(moves * "%")
            elif update_cmd[0] == "slotmove":
                moves = vardb.move_slot_ent(update_cmd, repo_match=repo_match)
                if moves:
                    writemsg_stdout(moves * "s")
                if bindb:
                    moves = bindb.move_slot_ent(update_cmd, repo_match=repo_match)
                    if moves:
                        writemsg_stdout(moves * "S")

    if world_modified:
        world_list.sort()
        write_atomic(world_file, "".join(f"{x}\n" for x in world_list))
        if world_warnings:
            # XXX: print warning that we've updated world entries
            # and the old name still matches something (from an overlay)?
            pass

    if retupd:

        def _config_repo_match(repo_name, atoma, atomb):
            """
            Check whether to perform a world change from atoma to atomb.
            If best vardb match for atoma comes from the same repository
            as the update file, allow that. Additionally, if portdb still
            can find a match for old atom name, warn about that.
            """
            matches = vardb.match(atoma)
            if not matches:
                matches = vardb.match(atomb)
                if not matches:
                    return False
            repository = vardb.aux_get(best(matches), ["repository"])[0]
            return repository == repo_name or (
                repo_name == master_repo and repository not in repo_map
            )

        update_config_files(
            root,
            shlex_split(mysettings.get("CONFIG_PROTECT", "")),
            shlex_split(mysettings.get("CONFIG_PROTECT_MASK", "")),
            repo_map,
            match_callback=_config_repo_match,
            case_insensitive="case-insensitive-fs" in mysettings.features,
        )

        # The above global updates proceed quickly, so they
        # are considered a single mtimedb transaction.
        if timestamps:
            # We do not update the mtime in the mtimedb
            # until after _all_ of the above updates have
            # been processed because the mtimedb will
            # automatically commit when killed by ctrl C.
            prev_mtimes.update(timestamps)

        # We gotta do the brute force updates for these now.
        def onUpdate(_maxval, curval):
            if curval > 0:
                writemsg_stdout("#")

        if quiet:
            onUpdate = None

        vardb.update_ents(repo_map, onUpdate=onUpdate)

        if bindb:

            def onUpdate(_maxval, curval):
                if curval > 0:
                    writemsg_stdout("*")

            bindb.update_ents(repo_map, onUpdate=onUpdate)

        # Update progress above is indicated by characters written to stdout so
        # we print a couple new lines here to separate the progress output from
        # what follows.
        writemsg_stdout("\n\n")

    return retupd



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["getmaskingreason"]

import portage
from portage import os
from portage.const import USER_CONFIG_PATH
from portage.dep import Atom, match_from_list
from portage.exception import InvalidAtom
from portage.localization import _
from portage.repository.config import _gen_valid_repo
from portage.util import grablines
from portage.versions import catpkgsplit, _pkg_str


def getmaskingreason(
    mycpv, metadata=None, settings=None, portdb=None, return_location=False, myrepo=None
):
    """
    If specified, the myrepo argument is assumed to be valid. This
    should be a safe assumption since portdbapi methods always
    return valid repo names and valid "repository" metadata from
    aux_get.
    """
    if settings is None:
        settings = portage.settings
    if portdb is None:
        portdb = portage.portdb
    mysplit = catpkgsplit(mycpv)
    if not mysplit:
        raise ValueError(_("invalid CPV: %s") % mycpv)

    if metadata is None:
        db_keys = list(portdb._aux_cache_keys)
        try:
            metadata = dict(zip(db_keys, portdb.aux_get(mycpv, db_keys, myrepo=myrepo)))
        except KeyError:
            if not portdb.cpv_exists(mycpv):
                raise
        else:
            if myrepo is None:
                myrepo = _gen_valid_repo(metadata["repository"])

    elif myrepo is None:
        myrepo = metadata.get("repository")
        if myrepo is not None:
            myrepo = _gen_valid_repo(metadata["repository"])

    if metadata is not None and not portage.eapi_is_supported(metadata["EAPI"]):
        # Return early since otherwise we might produce invalid
        # results given that the EAPI is not supported. Also,
        # metadata is mostly useless in this case since it doesn't
        # contain essential things like SLOT.
        if return_location:
            return (None, None)
        return None

    # Sometimes we can't access SLOT or repository due to corruption.
    pkg = mycpv
    try:
        pkg.slot
    except AttributeError:
        pkg = _pkg_str(mycpv, metadata=metadata, repo=myrepo)

    cpv_slot_list = [pkg]

    mycp = pkg.cp

    locations = []
    if pkg.repo in settings.repositories:
        for repo in settings.repositories[pkg.repo].masters + (
            settings.repositories[pkg.repo],
        ):
            locations.append(os.path.join(repo.location, "profiles"))
    locations.extend(settings.profiles)
    locations.append(os.path.join(settings["PORTAGE_CONFIGROOT"], USER_CONFIG_PATH))
    locations.reverse()
    pmasklists = []
    for profile in locations:
        pmask_filename = os.path.join(profile, "package.mask")
        node = None
        for l, recursive_filename in grablines(
            pmask_filename, recursive=1, remember_source_file=True
        ):
            if node is None or node[0] != recursive_filename:
                node = (recursive_filename, [])
                pmasklists.append(node)
            node[1].append(l)

    pmaskdict = settings._mask_manager._pmaskdict
    if mycp in pmaskdict:
        for x in pmaskdict[mycp]:
            if match_from_list(x, cpv_slot_list):
                x = x.without_repo
                for pmask in pmasklists:
                    comment = ""
                    comment_valid = -1
                    pmask_filename = pmask[0]
                    for i in range(len(pmask[1])):
                        l = pmask[1][i].strip()
                        try:
                            l_atom = Atom(
                                l, allow_repo=True, allow_wildcard=True
                            ).without_repo
                        except InvalidAtom:
                            l_atom = None
                        if l == "":
                            comment = ""
                            comment_valid = -1
                        elif l[0] == "#":
                            comment += l + "\n"
                            comment_valid = i + 1
                        elif l_atom == x:
                            if comment_valid != i:
                                comment = ""
                            if return_location:
                                return (comment, pmask_filename)
                            return comment
                        elif comment_valid != -1:
                            # Apparently this comment applies to multiple masks, so
                            # it remains valid until a blank line is encountered.
                            comment_valid += 1
    if return_location:
        return (None, None)
    return None



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import textwrap

import portage
from portage.dep import _repo_separator
from portage.elog import elog_process
from portage.elog.messages import eerror


def eapi_invalid(self, cpv, repo_name, settings, eapi_var, eapi_parsed, eapi_lineno):
    msg = []
    msg.extend(
        textwrap.wrap(
            (
                "EAPI assignment in ebuild '%s%s%s' does not"
                " conform with PMS section 7.3.1 (see bug #402167):"
            )
            % (cpv, _repo_separator, repo_name),
            70,
        )
    )

    if not eapi_parsed:
        # None means the assignment was not found, while an
        # empty string indicates an (invalid) empty assingment.
        msg.append(
            f"\tvalid EAPI assignment must occur on or before line: {eapi_lineno}"
        )
    else:
        msg.append(
            ("\tbash returned EAPI '%s' which does not match " "assignment on line: %s")
            % (eapi_var, eapi_lineno)
        )

    if portage.data.secpass >= 2:
        # TODO: improve elog permission error handling (bug #416231)
        for line in msg:
            eerror(line, phase="other", key=cpv)
        elog_process(cpv, settings, phasefilter=("other",))

    else:
        out = portage.output.EOutput()
        for line in msg:
            out.eerror(line)



# Copyright 2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class InstallKernelTestCase(TestCase):
    def testInstallKernel(self):
        ebuilds = {
            "sys-kernel/installkernel-systemd-boot-1": {
                "EAPI": "8",
                "RDEPEND": "!sys-kernel/installkernel-gentoo",
            },
            "sys-kernel/installkernel-gentoo-3": {
                "EAPI": "8",
                "RDEPEND": "!sys-kernel/installkernel-systemd-boot",
            },
            "sys-kernel/gentoo-kernel-5.15.23": {
                "EAPI": "8",
                "PDEPEND": ">=virtual/dist-kernel-5.15.23",
                "RDEPEND": "|| ( sys-kernel/installkernel-gentoo sys-kernel/installkernel-systemd-boot )",
            },
            "sys-kernel/gentoo-kernel-bin-5.15.23": {
                "EAPI": "8",
                "PDEPEND": ">=virtual/dist-kernel-5.15.23",
                "RDEPEND": "|| ( sys-kernel/installkernel-gentoo sys-kernel/installkernel-systemd-boot )",
            },
            "virtual/dist-kernel-5.15.23": {
                "EAPI": "8",
                "PDEPEND": "|| ( ~sys-kernel/gentoo-kernel-5.15.23 ~sys-kernel/gentoo-kernel-bin-5.15.23 )",
            },
        }

        installed = {
            "sys-kernel/installkernel-gentoo-3": {
                "EAPI": "8",
                "RDEPEND": "!sys-kernel/installkernel-systemd-boot",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                [
                    "sys-kernel/installkernel-systemd-boot",
                ],
                ambiguous_merge_order=True,
                success=True,
                mergelist=[
                    "sys-kernel/installkernel-systemd-boot-1",
                    "[uninstall]sys-kernel/installkernel-gentoo-3",
                    (
                        "!sys-kernel/installkernel-gentoo",
                        "!sys-kernel/installkernel-systemd-boot",
                    ),
                ],
            ),
            # Test bug 833014, where the calculation failed unless
            # --update and --deep are specified.
            ResolverPlaygroundTestCase(
                [
                    "sys-kernel/installkernel-systemd-boot",
                    "sys-kernel/gentoo-kernel-bin",
                ],
                ambiguous_merge_order=True,
                success=True,
                mergelist=[
                    "virtual/dist-kernel-5.15.23",
                    "sys-kernel/installkernel-systemd-boot-1",
                    "sys-kernel/gentoo-kernel-bin-5.15.23",
                    "[uninstall]sys-kernel/installkernel-gentoo-3",
                    (
                        "!sys-kernel/installkernel-systemd-boot",
                        "!sys-kernel/installkernel-gentoo",
                    ),
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2016 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class RuntimeCycleMergeOrderTestCase(TestCase):
    def testRuntimeCycleMergeOrder(self):
        ebuilds = {
            "app-misc/plugins-consumer-1": {
                "EAPI": "6",
                "DEPEND": "app-misc/plugin-b:=",
                "RDEPEND": "app-misc/plugin-b:=",
            },
            "app-misc/plugin-b-1": {
                "EAPI": "6",
                "RDEPEND": "app-misc/runtime-cycle-b",
                "PDEPEND": "app-misc/plugins-consumer",
            },
            "app-misc/runtime-cycle-b-1": {
                "RDEPEND": "app-misc/plugin-b app-misc/branch-b",
            },
            "app-misc/branch-b-1": {
                "RDEPEND": "app-misc/leaf-b app-misc/branch-c",
            },
            "app-misc/leaf-b-1": {},
            "app-misc/branch-c-1": {
                "RDEPEND": "app-misc/runtime-cycle-c app-misc/runtime-c",
            },
            "app-misc/runtime-cycle-c-1": {
                "RDEPEND": "app-misc/branch-c",
            },
            "app-misc/runtime-c-1": {
                "RDEPEND": "app-misc/branch-d",
            },
            "app-misc/branch-d-1": {
                "RDEPEND": "app-misc/leaf-d app-misc/branch-e",
            },
            "app-misc/branch-e-1": {
                "RDEPEND": "app-misc/leaf-e",
            },
            "app-misc/leaf-d-1": {},
            "app-misc/leaf-e-1": {},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/plugin-b"],
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    ("app-misc/leaf-b-1", "app-misc/leaf-d-1", "app-misc/leaf-e-1"),
                    ("app-misc/branch-d-1", "app-misc/branch-e-1"),
                    "app-misc/runtime-c-1",
                    ("app-misc/runtime-cycle-c-1", "app-misc/branch-c-1"),
                    "app-misc/branch-b-1",
                    ("app-misc/runtime-cycle-b-1", "app-misc/plugin-b-1"),
                    "app-misc/plugins-consumer-1",
                ],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2014-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import datetime
import subprocess
import sys
import textwrap

import portage
from portage import os, shutil, _shell_quote
from portage import _unicode_decode
from portage.const import PORTAGE_PYM_PATH, TIMESTAMP_FORMAT
from portage.process import find_binary
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class SyncLocalTestCase(TestCase):
    """
    Test sync with rsync and git, using file:// sync-uri.
    """

    def _must_skip(self):
        if find_binary("rsync") is None:
            self.skipTest("rsync: command not found")
        if find_binary("git") is None:
            self.skipTest("git: command not found")

    def testSyncLocal(self):
        debug = False
        self._must_skip()

        repos_conf = textwrap.dedent(
            """
			[DEFAULT]
			%(default_keys)s
			[test_repo]
			location = %(EPREFIX)s/var/repositories/test_repo
			sync-type = %(sync-type)s
			sync-depth = %(sync-depth)s
			sync-uri = file://%(EPREFIX)s/var/repositories/test_repo_sync
			sync-rcu = %(sync-rcu)s
			sync-rcu-store-dir = %(EPREFIX)s/var/repositories/test_repo_rcu_storedir
			auto-sync = %(auto-sync)s
			%(repo_extra_keys)s
		"""
        )

        profile = {"eapi": ("5",), "package.use.stable.mask": ("dev-libs/A flag",)}

        ebuilds = {
            "dev-libs/A-0": {},
            "sys-apps/portage-3.0": {"IUSE": "+python_targets_python3_8"},
        }

        installed = {
            "sys-apps/portage-2.3.99": {
                "EAPI": "7",
                "IUSE": "+python_targets_python3_8",
                "USE": "python_targets_python3_8",
            },
        }

        user_config = {"make.conf": ('FEATURES="metadata-transfer"',)}

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            profile=profile,
            user_config=user_config,
            debug=debug,
        )
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        homedir = os.path.join(eroot, "home")
        distdir = os.path.join(eprefix, "distdir")
        repo = settings.repositories["test_repo"]
        metadata_dir = os.path.join(repo.location, "metadata")
        rcu_store_dir = os.path.join(eprefix, "var/repositories/test_repo_rcu_storedir")

        cmds = {}
        for cmd in ("emerge", "emaint"):
            for bindir in (self.bindir, self.sbindir):
                path = os.path.join(bindir, cmd)
                if os.path.exists(path):
                    cmds[cmd] = (portage._python_interpreter, "-b", "-Wd", path)
                    break
            else:
                raise AssertionError(
                    f"{cmd} binary not found in {self.bindir} or {self.sbindir}"
                )

        git_binary = find_binary("git")
        git_cmd = (git_binary,)

        hg_binary = find_binary("hg")
        hg_cmd = (hg_binary,)

        committer_name = "Gentoo Dev"
        committer_email = "gentoo-dev@gentoo.org"

        def repos_set_conf(
            sync_type,
            dflt_keys=None,
            xtra_keys=None,
            auto_sync="yes",
            sync_rcu=False,
            sync_depth=None,
        ):
            env["PORTAGE_REPOSITORIES"] = repos_conf % {
                "EPREFIX": eprefix,
                "sync-type": sync_type,
                "sync-depth": 0 if sync_depth is None else sync_depth,
                "sync-rcu": "yes" if sync_rcu else "no",
                "auto-sync": auto_sync,
                "default_keys": "" if dflt_keys is None else dflt_keys,
                "repo_extra_keys": "" if xtra_keys is None else xtra_keys,
            }

        def alter_ebuild():
            with open(
                os.path.join(repo.location + "_sync", "dev-libs", "A", "A-0.ebuild"),
                "a",
            ) as f:
                f.write("\n")
            bump_timestamp()

        def bump_timestamp():
            bump_timestamp.timestamp += datetime.timedelta(seconds=1)
            with open(
                os.path.join(repo.location + "_sync", "metadata", "timestamp.chk"), "w"
            ) as f:
                f.write(
                    bump_timestamp.timestamp.strftime(
                        f"{TIMESTAMP_FORMAT}\n",
                    )
                )

        bump_timestamp.timestamp = datetime.datetime.utcnow()

        bump_timestamp_cmds = ((homedir, bump_timestamp),)

        sync_cmds = (
            (homedir, cmds["emerge"] + ("--sync",)),
            (
                homedir,
                lambda: self.assertTrue(
                    os.path.exists(os.path.join(repo.location, "dev-libs", "A")),
                    "dev-libs/A expected, but missing",
                ),
            ),
            (homedir, cmds["emaint"] + ("sync", "-A")),
        )

        sync_cmds_auto_sync = (
            (homedir, lambda: repos_set_conf("rsync", auto_sync="no")),
            (homedir, cmds["emerge"] + ("--sync",)),
            (
                homedir,
                lambda: self.assertFalse(
                    os.path.exists(os.path.join(repo.location, "dev-libs", "A")),
                    "dev-libs/A found, expected missing",
                ),
            ),
            (homedir, lambda: repos_set_conf("rsync", auto_sync="yes")),
        )

        rename_repo = (
            (homedir, lambda: os.rename(repo.location, repo.location + "_sync")),
        )

        rsync_opts_repos = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    None,
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % _shell_quote(repo.location + "_back"),
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (homedir, lambda: self.assertTrue(os.path.exists(repo.location + "_back"))),
            (homedir, lambda: shutil.rmtree(repo.location + "_back")),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        rsync_opts_repos_default = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % _shell_quote(repo.location + "_back"),
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (homedir, lambda: self.assertTrue(os.path.exists(repo.location + "_back"))),
            (homedir, lambda: shutil.rmtree(repo.location + "_back")),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        rsync_opts_repos_default_ovr = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % _shell_quote(repo.location + "_back_nowhere"),
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % _shell_quote(repo.location + "_back"),
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (homedir, lambda: self.assertTrue(os.path.exists(repo.location + "_back"))),
            (homedir, lambda: shutil.rmtree(repo.location + "_back")),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        rsync_opts_repos_default_cancel = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % _shell_quote(repo.location + "_back_nowhere"),
                    "sync-rsync-extra-opts = ",
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (
                homedir,
                lambda: self.assertFalse(os.path.exists(repo.location + "_back")),
            ),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        delete_repo_location = (
            (homedir, lambda: shutil.rmtree(repo.user_location)),
            (homedir, lambda: os.mkdir(repo.user_location)),
        )

        delete_rcu_store_dir = ((homedir, lambda: shutil.rmtree(rcu_store_dir)),)

        revert_rcu_layout = (
            (
                homedir,
                lambda: os.rename(repo.user_location, repo.user_location + ".bak"),
            ),
            (
                homedir,
                lambda: os.rename(
                    os.path.realpath(repo.user_location + ".bak"), repo.user_location
                ),
            ),
            (homedir, lambda: os.unlink(repo.user_location + ".bak")),
            (homedir, lambda: shutil.rmtree(repo.user_location + "_rcu_storedir")),
        )

        upstream_git_commit = (
            (
                repo.location + "_sync",
                git_cmd + ("commit", "--allow-empty", "-m", "test empty commit"),
            ),
            (
                repo.location + "_sync",
                git_cmd + ("commit", "--allow-empty", "-m", "test empty commit 2"),
            ),
        )

        delete_sync_repo = ((homedir, lambda: shutil.rmtree(repo.location + "_sync")),)

        git_repo_create = (
            (
                repo.location,
                git_cmd
                + (
                    "config",
                    "--global",
                    "user.name",
                    committer_name,
                ),
            ),
            (
                repo.location,
                git_cmd
                + (
                    "config",
                    "--global",
                    "user.email",
                    committer_email,
                ),
            ),
            (repo.location, git_cmd + ("init-db",)),
            (repo.location, git_cmd + ("add", ".")),
            (repo.location, git_cmd + ("commit", "-a", "-m", "add whole repo")),
        )

        sync_type_git = ((homedir, lambda: repos_set_conf("git")),)

        sync_type_git_shallow = (
            (homedir, lambda: repos_set_conf("git", sync_depth=1)),
        )

        sync_rsync_rcu = ((homedir, lambda: repos_set_conf("rsync", sync_rcu=True)),)

        delete_git_dir = (
            (homedir, lambda: shutil.rmtree(os.path.join(repo.location, ".git"))),
        )

        def hg_init_global_config():
            with open(os.path.join(homedir, ".hgrc"), "w") as f:
                f.write(f"[ui]\nusername = {committer_name} <{committer_email}>\n")

        hg_repo_create = (
            (repo.location, hg_init_global_config),
            (repo.location, hg_cmd + ("init",)),
            (repo.location, hg_cmd + ("add", ".")),
            (repo.location, hg_cmd + ("commit", "-A", "-m", "add whole repo")),
        )

        sync_type_mercurial = ((homedir, lambda: repos_set_conf("mercurial")),)

        def append_newline(path):
            with open(path, "a") as f:
                f.write("\n")

        upstream_hg_commit = (
            (
                repo.location + "_sync",
                lambda: append_newline(
                    os.path.join(repo.location + "_sync", "metadata/layout.conf")
                ),
            ),
            (
                repo.location + "_sync",
                hg_cmd + ("commit", "metadata/layout.conf", "-m", "test empty commit"),
            ),
            (
                repo.location + "_sync",
                lambda: append_newline(
                    os.path.join(repo.location + "_sync", "metadata/layout.conf")
                ),
            ),
            (
                repo.location + "_sync",
                hg_cmd
                + ("commit", "metadata/layout.conf", "-m", "test empty commit 2"),
            ),
        )

        if hg_binary is None:
            mercurial_tests = ()
        else:
            mercurial_tests = (
                delete_sync_repo
                + delete_git_dir
                + hg_repo_create
                + sync_type_mercurial
                + rename_repo
                + sync_cmds
                + upstream_hg_commit
                + sync_cmds
            )

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "DISTDIR": distdir,
            "GENTOO_COMMITTER_NAME": committer_name,
            "GENTOO_COMMITTER_EMAIL": committer_email,
            "HOME": homedir,
            "PATH": os.environ["PATH"],
            "PORTAGE_GRPNAME": os.environ["PORTAGE_GRPNAME"],
            "PORTAGE_USERNAME": os.environ["PORTAGE_USERNAME"],
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
        }
        repos_set_conf("rsync")

        if os.environ.get("SANDBOX_ON") == "1":
            # avoid problems from nested sandbox instances
            env["FEATURES"] = "-sandbox -usersandbox"

        dirs = [homedir, metadata_dir]
        try:
            for d in dirs:
                ensure_dirs(d)

            timestamp_path = os.path.join(metadata_dir, "timestamp.chk")
            with open(timestamp_path, "w") as f:
                f.write(
                    bump_timestamp.timestamp.strftime(
                        f"{TIMESTAMP_FORMAT}\n",
                    )
                )

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for cwd, cmd in (
                rename_repo
                + sync_cmds_auto_sync
                + sync_cmds
                + rsync_opts_repos
                + rsync_opts_repos_default
                + rsync_opts_repos_default_ovr
                + rsync_opts_repos_default_cancel
                + bump_timestamp_cmds
                + sync_rsync_rcu
                + sync_cmds
                + delete_rcu_store_dir
                + sync_cmds
                + revert_rcu_layout
                + delete_repo_location
                + sync_cmds
                + sync_cmds
                + bump_timestamp_cmds
                + sync_cmds
                + revert_rcu_layout
                + delete_sync_repo
                + git_repo_create
                + sync_type_git
                + rename_repo
                + sync_cmds
                + upstream_git_commit
                + sync_cmds
                + sync_type_git_shallow
                + upstream_git_commit
                + sync_cmds
                + mercurial_tests
            ):
                if hasattr(cmd, "__call__"):
                    cmd()
                    continue

                abs_cwd = os.path.join(repo.location, cwd)
                proc = subprocess.Popen(cmd, cwd=abs_cwd, env=env, stdout=stdout)

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != os.EX_OK:
                        for line in output:
                            sys.stderr.write(_unicode_decode(line))

                self.assertEqual(
                    os.EX_OK,
                    proc.returncode,
                    f"{cmd} failed in {cwd}",
                )

        finally:
            playground.cleanup()



# Copyright 2012-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import re
import subprocess
import sys

from portage import _unicode_encode, _unicode_decode
from portage.util import writemsg
from portage.util.configparser import RawConfigParser, read_configs


def parse_desktop_entry(path):
    """
    Parse the given file with RawConfigParser and return the
    result. This may raise an IOError from io.open(), or a
    ParsingError from RawConfigParser.
    """
    parser = RawConfigParser()

    read_configs(parser, [path])

    return parser


_trivial_warnings = re.compile(
    r" looks "
    # >=desktop-file-utils-0.25
    r"(?:the same as that of key|"
    # <desktop-file-utils-0.25
    r"redundant with value) "
)

_ignored_errors = (
    # Ignore error for emacs.desktop:
    # https://bugs.freedesktop.org/show_bug.cgi?id=35844#c6
    'error: (will be fatal in the future): value "TextEditor" in key "Categories" in group "Desktop Entry" requires another category to be present among the following categories: Utility',
    'warning: key "Encoding" in group "Desktop Entry" is deprecated',
)

_ShowIn_exemptions = (
    # See bug #480586.
    'contains an unregistered value "Pantheon"',
)


def validate_desktop_entry(path):
    args = ["desktop-file-validate", path]

    args = [_unicode_encode(x, errors="strict") for x in args]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output_lines = _unicode_decode(proc.communicate()[0]).splitlines()
    proc.wait()

    if output_lines:
        filtered_output = []
        for line in output_lines:
            msg = line[len(path) + 2 :]
            # "hint:" output is new in desktop-file-utils-0.21
            if msg.startswith("hint: ") or msg in _ignored_errors:
                continue
            if (
                'for key "NotShowIn" in group "Desktop Entry"' in msg
                or 'for key "OnlyShowIn" in group "Desktop Entry"' in msg
            ):
                exempt = False
                for s in _ShowIn_exemptions:
                    if s in msg:
                        exempt = True
                        break
                if exempt:
                    continue
            filtered_output.append(line)
        output_lines = filtered_output

    if output_lines:
        output_lines = [
            line for line in output_lines if _trivial_warnings.search(line) is None
        ]

    return output_lines


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        for line in validate_desktop_entry(arg):
            writemsg(line + "\n", noiselevel=-1)



# Copyright 2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os


def installed_dynlibs(directory):
    for _dirpath, _dirnames, filenames in os.walk(directory):
        for filename in filenames:
            if filename.endswith(".so"):
                return True
    return False


def check_dyn_libs_inconsistent(directory, provides):
    """Checks directory for whether any dynamic libraries were installed and
    if PROVIDES corresponds."""

    # Let's check if we've got inconsistent results.
    # If we're installing dynamic libraries (.so files), we should
    # really have a PROVIDES.
    # (This is a complementary check at the point of ingestion for the
    # creation check in doebuild.py)
    # Note: we could check a non-empty PROVIDES against the list of .sos,
    # but this doesn't gain us anything. We're interested in failure
    # to properly parse the installed files at all, which should really
    # be a global problem (e.g. bug #811462)
    return not provides and installed_dynlibs(directory)
